!include "MUI2.nsh"
!include "LogicLib.nsh"

;Unicode true

;SetCompress off
SetCompressor zlib
;SetCompressor /solid LZMA
;SetCompressorDictSize 8
;RequestExecutionLevel highest
RequestExecutionLevel admin

;!define MUI_ICON "${NSISDIR}\Contrib\Graphics\Icons\nsis3-install-alt.ico"
;!define MUI_UNICON "${NSISDIR}\Contrib\Graphics\Icons\nsis3-uninstall.ico"
;!define MUI_WELCOMEFINISHPAGE_BITMAP "${NSISDIR}\Contrib\Graphics\Wizard\nsis3-metro.bmp"
!define MUI_ICON "${NSISDIR}\Contrib\Graphics\Icons\orange-install.ico"
!define MUI_UNICON "${NSISDIR}\Contrib\Graphics\Icons\orange-uninstall.ico"
!define MUI_WELCOMEFINISHPAGE_BITMAP "${NSISDIR}\Contrib\Graphics\Wizard\orange.bmp"
!define MUI_COMPONENTSPAGE_NODESC
!define MUI_FINISHPAGE_NOAUTOCLOSE

!searchparse /file ..\src\euterpe\app_info.h '#define EUTERPE_MAJOR_VERSION ' MAJOR
!searchparse /file ..\src\euterpe\app_info.h '#define EUTERPE_MINOR_VERSION ' MINOR
!searchparse /file ..\src\euterpe\app_info.h '#define EUTERPE_RELEASE "' RELEASE '"'
!searchparse /file ..\src\euterpe\app_info.h '#define EUTERPE_COPYRIGHT "' COPYRIGHT '"'
!searchparse /file ..\src\euterpe\app_info.h '#define EUTERPE_COMPANY_URL "' COMPANY '"'

Name "Euterpe"
OutFile "euterpe-${MAJOR}.${MINOR}${RELEASE}-x86-install.exe"
;BrandingText "Grafmin"

VIAddVersionKey "ProductName" "Euterpe"
VIAddVersionKey "CompanyName" "${COMPANY}"
VIAddVersionKey "FileDescription" "Euterpe installer (x86)"
VIAddVersionKey "FileVersion" "${MAJOR}.${MINOR}.0.0"
VIAddVersionKey "ProductVersion" "${MAJOR}.${MINOR}"
VIAddVersionKey "LegalCopyright" "${COPYRIGHT}"
VIProductVersion "${MAJOR}.${MINOR}.0.0"

InstallDir "$PROGRAMFILES32\Euterpe"

;Get installation folder from registry if available
InstallDirRegKey HKCU "Software\Euterpe (x86)" ""

!define MUI_ABORTWARNING
!define MUI_FINISHPAGE_RUN $INSTDIR\euterpe.exe

!insertmacro MUI_PAGE_LICENSE "..\doc\license.txt"
!insertmacro MUI_PAGE_COMPONENTS
!insertmacro MUI_PAGE_DIRECTORY
!insertmacro MUI_PAGE_INSTFILES
!insertmacro MUI_PAGE_FINISH
;AutoCloseWindow true
 
!insertmacro MUI_UNPAGE_CONFIRM
!insertmacro MUI_UNPAGE_INSTFILES

!insertmacro MUI_LANGUAGE "English"

Section "Euterpe" SecEuterpe
  SectionIn RO

  SetOutPath "$INSTDIR"
  
  File "..\build\vc\Win32\Release\euterpe.exe"
  File "..\build\vc\Win32\Release\*.dll"
  File "..\build\vc\Win32\Release\changelog.txt"
  File "..\build\vc\Win32\Release\license.txt"
  File "..\build\vc\Win32\Release\documentation.txt"
  
  SetOutPath "$INSTDIR\plugins"
  File /r "..\build\vc\Win32\Release\plugins\*.dll"

  SetOutPath "$INSTDIR\themes"
  File /r "..\build\vc\Win32\Release\themes\*.*"
 
  ;Store installation folder
  WriteRegStr HKCU "Software\Euterpe (x86)" "" $INSTDIR
  
  ;Create uninstaller
  WriteUninstaller "$INSTDIR\uninstall.exe"
SectionEnd

Section "Start menu shortcuts" SecStart
  SetOutPath "$SMPROGRAMS\Euterpe (x86)"
  CreateShortcut "$OUTDIR\Euterpe (x86).lnk" "$INSTDIR\euterpe.exe"
  CreateShortcut "$OUTDIR\Uninstall Euterpe (x86).lnk" "$INSTDIR\uninstall.exe"
SectionEnd

;Section "Context menu entry (File Explorer)" SecContext
;  WriteRegStr HKCR "*\shell\Euterpe (x86)" "" "Send in &Euterpe (x86)"
;  WriteRegStr HKCR "*\shell\Euterpe (x86)\command" "" '"$INSTDIR\euterpe.exe" --add "%1"'
;SectionEnd

Section /o "Desktop icon" SecDesktop
  SetOutPath $DESKTOP
  CreateShortcut "$OUTDIR\Euterpe (x86).lnk" "$INSTDIR\euterpe.exe"
SectionEnd

;Language strings
LangString DESC_SecEuterpe ${LANG_ENGLISH} "Euterpe core (required)."
LangString DESC_SecStart ${LANG_ENGLISH} "Start menu shortcuts."
LangString DESC_SecDesktop ${LANG_ENGLISH} "Desktop icon."

;Assign language strings to sections
!insertmacro MUI_FUNCTION_DESCRIPTION_BEGIN
!insertmacro MUI_DESCRIPTION_TEXT ${SecEuterpe} $(DESC_SecEuterpe)
!insertmacro MUI_DESCRIPTION_TEXT ${SecStart} $(DESC_SecStart)
!insertmacro MUI_DESCRIPTION_TEXT ${SecDesktop} $(DESC_SecDesktop)
!insertmacro MUI_FUNCTION_DESCRIPTION_END

;--------------------------------
;Uninstaller Section

Section "Uninstall"
  StrCpy $1 "euterpe.exe"
  
retry:
  nsProcess::_FindProcess "$1"
  
  Pop $R0
  ${if} $R0 = 0
    MessageBox MB_RETRYCANCEL "Euterpe is running. Close Euterpe and hit Retry." /SD IDRETRY IDCANCEL cancel
    Goto retry
  ${endif}

  Delete "$DESKTOP\Euterpe (x86).lnk"
  Delete "$SMPROGRAMS\Euterpe (x86)\Euterpe (x86).lnk"
  Delete "$SMPROGRAMS\Euterpe (x86)\Uninstall Euterpe (x86).lnk"
  RMDir /r "$SMPROGRAMS\Euterpe (x86)"

  Delete "$INSTDIR\euterpe.exe"
  Delete "$INSTDIR\*.dll"
  Delete "$INSTDIR\changelog.txt"
  Delete "$INSTDIR\license.txt"
  Delete "$INSTDIR\documentation.txt"
  RMDir /r "$INSTDIR\plugins"
  RMDir /r "$INSTDIR\themes"
  Delete "$INSTDIR\uninstall.exe"
  RMDir "$INSTDIR"

  DeleteRegKey /ifempty HKCU "Software\Euterpe (x86)"
  ;DeleteRegKey HKCR "*\shell\Euterpe (x86)"
  
  Goto cont
  
cancel:
  Abort
  
cont:
SectionEnd
